#!/usr/bin/env python

# script to parse a file that was created via copy/paste of an iscope workstation
# search result.  We will read each line, pull tokens, and output in a format suitable
# for a regular expression, as part of ABA integration
#
# TYPICAL METRIC:: usilap533c	WebSphere	CCM02A	Backends|prt1 on rtcprod.ca.com-12000 (DB2 DB)|SQL|Prepared|Query	Average Response Time (ms)	60.0	0.0	18777.0	228580
#                  ----0----    ---1----        --2---  ----------------------------------3------------------------------       ----------4---------------      > DON'T CARE------------------
#
# SYSVIEW metrics are derived from a different process and thus have a different output format.  These are identified
# by having 'sysview' appear in the filename that contains sysview candidate metrics.  These are being avoided for now.
#
# SYSVIEW METRIC:: CICS Regions|CACMPRO2	Average Lifetime Per Transaction (µs)	1.497638503E9	0.0	1.497638503E9	2015
#                  ---------0-----------        -----------------1-------------------   > DON'T CARE--------------------------------
#
# WHITESPACE metrics:  When a copy/paste of the search results ends up losing the '\t' delimiters, we split first on the metric unit/descriptor.
#  this is just a nightmare of splits and counting results... in order to decide parsing... ugh.
#
# msydor : 20141008 : Implemented -list/-file functionality
# msydor : 20140930 : Reorganized for an output function
#                   : added explicit regex syntax check before writing regex expression for ABA
# msydor : 20140923 : Added -list, -file and -index functionality
#                     (x) need to implement -list/-file
# msydor : 20140922 : changed metric regex from '+' to '\\|[^|]+' - but not sure I really need it...
#                   : corrected metric units duplication bug
#                   : put in a data structure to organize the metric parameters and simplify code...
#                   : corrected excess Stalled Metric bugs
# msydor : 20140916 : cleaned up tabbed metrics to align with whitespace metrics
# msydor : 20140915 : changed to argparse for input parameter management
# msydor : 20140910 : now handles exception tracking and less common metric units/types
# msydor : 20140908 : now works on a file containing a list of files to process 
# msydor : 20140903 : initial prototype
#
#

_BROAD  = 'broad'  # grab all metric types for a given metric
_NARROW = 'narrow' # use only the specific metric type presented
_NONE   = 'none'   # no debug
_LOW    = 'low'    # _DEBUG2
_HIGH   = 'high'   # _DEBUG
_ALL    = 'all'    # all DEBUG options...

import re
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('-list', help='supplying a list of filenames of candidate metrics to be processed into regex suitable for ABA')
parser.add_argument('-output', default='buildABA_regex', help='file to capture regex suitable for ABA')
parser.add_argument('-exceptions', default='buildABA_exceptions', help='file to capture any unexpected metric types')
parser.add_argument('-depth', choices =[_BROAD,_NARROW],default=_BROAD, help='controls quantity of regex (more/less metrics)')
parser.add_argument('-debug', choices =[_NONE,_LOW,_HIGH, _ALL],default=_NONE, help='controls debug info level into output file')
parser.add_argument('-file', default='kpis', help='supplying a single file of candidate metrics to process')
parser.add_argument('-index', type = int, default=0, help='set the starting index for the regular expressions')
_args = parser.parse_args()


_index = _args.index
_linesWritten = 0

_validCOUNT = 0
_errorCOUNT = 0
_metricCOUNT = 0

# ABA regex format in Analytics.properties
_expression = 'analytics.metricfeed.expr.'
_domain     = 'analytics.metricfeed.domain.'
_process    = 'analytics.metricfeed.process.'
_metric     = 'analytics.metricfeed.metric.'

_SYSVIEW_METRIC = False

_DEBUG = False
_DEBUG2 = False
if ( _args.debug == _HIGH):_DEBUG  = True      # controls overall debug information - verbose
if ( _args.debug == _LOW ):_DEBUG2 = True      # controls debug info appearing in output file for ABA configuration
if ( _args.debug == _ALL ):
    _DEBUG2 = True      
    _DEBUG  = True
    
_NO_TABS = True

###
### list of row lists - extend for new metric units here...
###  - make sure the 'label' is unique.  First match wins!
###
### This data structure holds the 'label' (unique identifer, used in debug), the metric string to match and a flag
### to show which metric is being used in each iteration
###
MetricUnit = []
MetricUnit.append ({'label':'AVG',              'metricString':'Average Response Time (ms)',          'present':False} )
MetricUnit.append ({'label':'AVGR',             'metricString':'Average Result Processing (ms)',      'present':False} )
MetricUnit.append ({'label':'AVG',              'metricString':'Average Response Time (ms)',          'present':False} )
MetricUnit.append ({'label':'AVGR',             'metricString':'Average Result Processing Time (ms)', 'present':False} )
MetricUnit.append ({'label':'AVGP',             'metricString':'Average Processing Time (ms)',        'present':False} )
MetricUnit.append ({'label':'AVGRPS',           'metricString':'Average Responses Per Second',        'present':False} )
MetricUnit.append ({'label':'MAFE',             'metricString':'Max Average Frontend Reponse Time ',  'present':False} )
MetricUnit.append ({'label':'MAVG',             'metricString':'Max Average Response Time',           'present':False} )
MetricUnit.append ({'label':'AVGMI',            'metricString':'Average Method Invocation Time (ms)', 'present':False} )
MetricUnit.append ({'label':'CON',              'metricString':'Concurrent Invocations',              'present':False} )
MetricUnit.append ({'label':'RES',              'metricString':'Responses Per Interval',              'present':False} )
MetricUnit.append ({'label':'STALLED',          'metricString':'number of stalled components',        'present':False} )
MetricUnit.append ({'label':'STALL',            'metricString':'Stall Count',                         'present':False} )
MetricUnit.append ({'label':'ERR',              'metricString':'Errors Per Interval',                 'present':False} )
## SAP Specific
MetricUnit.append ({'label':'HOSTAVGCPU',       'metricString':'Host Average CPU',                    'present':False} )
MetricUnit.append ({'label':'AVGWAIT',          'metricString':'Average Wait Time',                   'present':False} )
MetricUnit.append ({'label':'IODBWRITELATENCY', 'metricString':'I/O Database Writes Average Latency', 'present':False} )
MetricUnit.append ({'label':'IODBREADLATENCY',  'metricString':'I/O Database Reads Average Latency',  'present':False} )
MetricUnit.append ({'label':'NEM',              'metricString':'Number of Evaluated Metrics',         'present':False} )

# old hack to be resolved...
_MAFE = 'Max Average Frontend Reponse Time '
_MAFE_PRESENT = False
_MAVG = 'Max Average Response Time'
_MAVG_PRESENT = False

##_METRIC_MATCH = '+:.+'      # used for a 'broad' set of candidate metrics
_METRIC_MATCH = '\\\|[^|]+:.+'      # used for a 'broad' set of candidate metrics
##_METRIC_SPECIFIC = '+:'     # used for a 'narrow' set of candidate metrics
_METRIC_SPECIFIC = '\\\|[^|]+:'     # used for a 'narrow' set of candidate metrics
_3SPACE = '   '
_2SPACE = '  '
_VIRTUAL = '(Virtual)'
_RPAREN = ')'
_START_WITH_CUSTOM = False
_process_CUSTOM = 'Custom Metric Host (Virtual)\\|Custom Metric Process (Virtual)\\|Custom Business Application Agent (Virtual)'

# Exceptional Cases - hack to be reconsidered... was probably due to an earlier typo...
_JMS = 'JMS|'
_JMS_PRESENT = False

###
### MAIN
###

fileOUT = open(_args.output,'w')
print("...writing to ", _args.output)
fileEXCEPTIONS = open(_args.exceptions,'w')
print("...writing exceptions to ", _args.exceptions)

def getFile(fileArg, listArg):
    listOfFiles = []

    if ( listArg ):
        with open(listArg) as fileLIST:       
            for _lineFILE in fileLIST:
                listOfFiles.append(_lineFILE)
    else:
        listOfFiles.append( _args.file )

    return listOfFiles

def writeRegexProcess(process, metric, scope, index, lines, counts, valids, errors):
    myregex_p = process
    myregex = metric
    counts = counts + 1

    try:
        re.compile( myregex_p )
    except:
        errors = errors + 1
        print( '_regex_p::',myregex_p,'\n',sep='',file=fileEXCEPTIONS )
    else:
        try:
            re.compile( myregex )
        except:
            errors = errors + 1
            print( '_regex::',myregex,'\n',sep='',file=fileEXCEPTIONS )
        else:
            valids = valids + 1
            print (_expression,index,' = ',_domain,index,', ',_process,index,', ',_metric,index,sep='',file=fileOUT)
            print (_domain,index,' = ','.*',sep='',file=fileOUT)
            print (_process,index,' = ',myregex_p,sep='',file=fileOUT)
            print (_metric,index,' = ',myregex.rstrip(),scope,'\n\n', sep='',file=fileOUT)
            index = index + 1
            lines = lines + 6
            
    return index, lines, counts, valids, errors

def writeRegexMetric(metric, scope, index, lines, counts, valids, errors):
    myregex = metric
    counts = counts + 1
    
    try:
      re.compile( myregex )
    except:
        errors = errors + 1
        print( '_regex::',myregex,'\n',sep='',file=fileEXCEPTIONS )
    else:
        valids = valids + 1
        print (_expression,index,' = ',_domain,index,', ',_process,index,', ',_metric,index,sep='',file=fileOUT)
        print (_domain,index,' = ','.*',sep='',file=fileOUT)
        print (_process,index,' = ','.*',sep='',file=fileOUT)
        print (_metric,index,' = ',myregex.rstrip(),scope,'\n\n', sep='',file=fileOUT)
        index = index + 1
        lines = lines + 6
        
    return index, lines, counts, valids, errors


for _lineFILE in getFile(_args.file, _args.list):  # for each filename in fileLIST
    
    _input = _lineFILE.split("\n")  # get rid of the '\n'
    print("...processing ",_input,"\t Index::",_index,"\tLines Written::",_linesWritten)
    
    ##### SYSVIEW Considerations
    #####
    _test = _lineFILE[0].find("sysview")
    if (_DEBUG == True):print("=======",_input[0],"========",_index,sep='',file=fileOUT)
    
    if ((_test > 0) == True):
        _SYSVIEW_METRIC = True
        if(_DEBUG == True):print(">>> SYSVIEW file detected <<<",_index,sep='',file=fileOUT)
    else:
        _SYSVIEW_METRIC = False
        if(_DEBUG == True):print(">>> standard file <<<",_index,sep='',file=fileOUT)
        
    ##### Main Processing Loop
    #####
    with open(_input[0]) as fileIN:
        for _lineMETRIC in fileIN:  # for each metric in the file

            if (_lineMETRIC.startswith('Custom')>0):
                _START_WITH_CUSTOM = True  # Custom Metric Host
                if(_DEBUG == True): print(">>>Starts with Custom...",_index,sep='',file=fileOUT)
                if(_DEBUG2 == True): print(">>>Starts with Custom...","::",_index,sep='',file=fileOUT)
                
            if (_lineMETRIC.find('\t')>0):
                _NO_TABS = False  # Tab delimiter is in use...easy!
                if(_DEBUG == True): print(">>>Tab delimiter is in use...",_index,sep='',file=fileOUT)
            
        ### do test for metric type here... yielding _SCOPE
        #
        #
            _SCOPE = ''
            #
            # determine type of metric and make that split...
            #
            if(_SYSVIEW_METRIC == False):      # SYSVIEW metrics are fairly unique... and are not yet accomodated
                # get the appropriate units string...subject to the overall scope (broad/narrow)
                for metric in MetricUnit:
                    if ( _lineMETRIC.find( metric['metricString'] ) > 0 ):
                        metric['present'] = True
                        tmp1 = _lineMETRIC.split(metric['metricString'])
                        if (   _args.depth == _BROAD  ): _SCOPE = _METRIC_MATCH
                        elif ( _args.depth == _NARROW ): _SCOPE = _METRIC_SPECIFIC + metric['metricString']

                        # temporary patch...
                        if ( metric['metricString'] == _MAVG ): _MAVG_PRESENT = True
                        if ( metric['metricString'] == _MAFE ): _MAFE_PRESENT = True
                        
                        if (_DEBUG == True):
                            print(metric['label'],">> LEN::", len(tmp1)," SPLIT::", tmp1, " SCOPE::", _SCOPE," INDEX::",_index,sep='',file=fileOUT)
            else:
                # this is your invitation to add new metric units to the MetricUnit data structure.
                print("WARNING :: No valid metric unit (string) found for current capability.  Please consider adding to the current configuration.\n")
                print("_lineMETRIC::",_lineMETRIC,'\n')
                        
                
            if (_NO_TABS == False):
                tokens = _lineMETRIC.split("\t")
                
                if(_SYSVIEW_METRIC == True):
### need to account for broad/narrow with sysview metrics...
#
#
                    _SCOPE = 'SYSVIEW'  #  patch!!!
                            
                    # sysview metrics are present (special case, different file format)
                    if (len(tokens) >=2):
                        if (_DEBUG2 == True):print ("ORIG:a:",tokens,"::",_index,sep='',file=fileOUT)

                        _tmp5 = tokens.split(':')
                        if ( len( _tmp5[0] ) >5 ):    # a reasonable left hand term...

                            _index, _linesWritten = writeRegexProcess(tokens[0],token[1],_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    else:
                        print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                        print("TOKENS:a:",len(tokens),":",tokens,sep='',file=fileEXCEPTIONS)
                        print("sysview metrics are present (special case, different file format) ::",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)
                    
                else:
                    # sysview metrics are absent... (typical)

                    # format final results for extra '|'s
                    if ( len( tokens) > 3 ): tmpFormatted = tokens[3].replace('|','\\\|')
                        
                    if (len(tokens) > 3):
                        if (_START_WITH_CUSTOM == True):
                            if (_DEBUG == True):
                                print("... Custom Metric Host...", "Got ",len (tokens)," == ",tokens,_index,sep='',file=fileOUT)
                        
# more code here... for dealing with Custom Metric Host... need to change indexes....
#
#
                            if (_DEBUG == True): print(">>> got tokens == 4 ",tokens,_index,sep='',file=fileOUT)
                            if (_DEBUG == True): print(">::",_index,"::", tokens[1],"::", tokens[2],"::", tokens[4],_index,sep='',file=fileOUT)

                       
                            if (len(tokens) >= 4):
                                if (_DEBUG2 == True):print ("ORIG:b:",tokens,"::",_index,sep='',file=fileOUT)

                                _tmp5 = tmpFormatted.split(':')
                                if ( len( _tmp5[0] ) >5 ):    # a reasonable left hand term...
                                    tmp10 = tokens[1]+'\\\|'+tokens[2]
                                    _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexProcess(tmp10,tmpFormatted,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                            else:
                                print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                                print("TOKENS:b:",len(tokens),":",tokens,sep='',file=fileEXCEPTIONS)
                                print("tmpFORMATTED:b:",tmpFormatted,sep='',file=fileEXCEPTIONS)
                                print("dealing with Custom metrics ::",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)
                            
                        else:    # regular metrics...(not Custom)
                            if (_DEBUG == True):
                                print("... Regular Metric...", "Got ",len (tokens)," == ",tokens,_index,sep='',file=fileOUT)
                        
                            if (_DEBUG == True): print(">>> got tokens == 4 ",tokens,_index,sep='',file=fileOUT)
                            if (_DEBUG == True): print(">::",_index,"::", tokens[1],"::", tokens[2],"::", tokens[4],_index,sep='',file=fileOUT)
                       
                            
                            if (len(tokens) >= 4):
                                if (_DEBUG2 == True):print ("ORIG:c:",tokens,"::",_index,sep='',file=fileOUT)

                                
                                _tmp5 = tmpFormatted.split(':')
                                if ( len( _tmp5[0] ) >5 ):    # a reasonable left hand term...
                                    tmp10 = tokens[1]+'\\\|'+tokens[2]
                                    _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexProcess(tmp10,tmpFormatted,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                            else:
                                print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                                print("TOKENS:c:",len(tokens),":",tokens,sep='',file=fileEXCEPTIONS)
                                print("tmpFORMATTED:b:",tmpFormatted,sep='',file=fileEXCEPTIONS)
                                print("dealing with regular metrics ::",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)
                            
                    else:
                        
                        if (_DEBUG == True):
                            print("... bad parsing...", "Got ",len (tokens)," == ",tokens,_index,sep='',file=fileOUT)
                            print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                            print("TOKENS:x:",len(tokens),":",tokens,sep='',file=fileEXCEPTIONS)
                            print("fell through processing to ELSE clause... ::",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)
                            print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileOUT)
                            print("TOKENS:x:",len(tokens),":",tokens,sep='',file=fileOUT)
                            print("fell through processing to ELSE clause... ::",_linesWritten,"\n",sep='',file=fileOUT)

                
            else:    # no tabs - only whitespace delimited... need to use other features - this gets very fussy...
                    
                if(_DEBUG == True): print(">>>Tab delimiter NOT in use...",_index,sep='',file=fileOUT)
                
                # Special Cases - these have a newer parsing strategy  <<< this needs to be reviewed...
                if (_lineMETRIC.find(_JMS)>0):
                    _JMS_PRESENT = True
                    _JMS_temp = _lineMETRIC.split(_JMS)
                    _START_WITH_CUSTOM = False
                    if (_DEBUG == True): print("JMS in _lineMETRIC>> LEN::", len(_JMS_temp)," SPLIT::", _JMS_temp," INDEX::",_index,sep='',file=fileOUT)
                elif (tmp1[0].find(_JMS)> 0 ) :
                    _JMS_PRESENT = True
                    _JMS_temp = _lineMETRIC.split(_JMS)
                    _START_WITH_CUSTOM = False
                    if (_DEBUG == True): print("JMS in _tmp1[0]>> LEN::", len(_JMS_temp)," SPLIT::", _JMS_temp," INDEX::",_index,sep='',file=fileOUT)

                        
                #
                # dealing with the whitespace
                #
                
                if (_START_WITH_CUSTOM == True): # Custom Metrics
                    if (_DEBUG == True): print("... whitespace, custom", "Got ",len (tmp1)," == ",tmp1," INDEX::",_index,sep='',file=fileOUT)

                    #
                    # assemble the output
                    #
                    tmp3 = tmp1[0].split(_2SPACE)
                    tmp4 = tmp3[2].split(_VIRTUAL)
                    if (len(tmp4) >= 3):
                        tmp5 = tmp4[2].split(_RPAREN)
                    elif (len(tmp3) == 4):
                        tmp5 = tmp3[3].split(" ")
                    else:
                        print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                        print("ORIGINAL::",tmp1[0],sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp4:",len(tmp4),":",tmp4,sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp3:",len(tmp3),":",tmp3,sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp1:",len(tmp1),":",tmp1,sep='',file=fileEXCEPTIONS)
                        print("at Assembly, dealing with whitespace, Custom metrics :y:",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)
                        
                    if (_DEBUG == True): print("[][][] whitespace, custom::", "[---]",tmp3, "[---]",tmp4," INDEX::",_index,sep='',file=fileOUT)

                    if (_JMS_PRESENT == True):
                        # get the appropriate units string...
                        for metric in MetricUnit:
                            if ( metric['present'] == True):_tmp1 = _JMS_temp[1].split(metric['metricString'])
                        
                        _tmp2 = _JMS + _tmp1[0]
                        _tmp3 = _tmp2.rstrip()
                        _tmp4 = _tmp3.replace('|','\\\|')+_SCOPE
                        
                        if (_DEBUG2 == True):
                            print ("CUSTOM:ORIG:f3:",_JMS_temp[1],"::",_index,sep='',file=fileOUT)
                            print ("_lineMETRIC:f3:",_lineMETRIC,"::",_index,sep='',file=fileOUT)
                            print ("_tmp1:f3:",_tmp1,"::",_index,sep='',file=fileOUT)
                            print ("_tmp2:f3:",_tmp2,"::",_index,sep='',file=fileOUT)
                            print ("_tmp3:f3:",_tmp3,"::",_index,sep='',file=fileOUT)
                            print ("_tmp4:f3:",_tmp4,"::",_index,sep='',file=fileOUT)

                        _tmp5 = _tmp4.split(':')
                        if ( len( _tmp5[0] ) >5 ):    # a reasonable left hand term...
                            _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexProcess(_process_CUSTOM,_tmp4,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    elif ( _MAVG_PRESENT == True or _MAFE_PRESENT):  # improved parsing...  <<< old hack that should probably be eliminated...
                        if (_JMS_PRESENT == True):
                            # get the appropriate units string...
                            for metric in MetricUnit:
                                if ( metric['present'] == True):_tmp1 = _tmp2 = tmp1[0].split(metric['metricString'])

                        _tmp3 = _tmp2[0].split(_3SPACE)
                        if ( len ( _tmp3 ) < 3 ):
                            print('Parsing Error, expected 3 items')
                            print('_3SPACE applied to::',_tmp2[0])
                            print('results in::',_tmp3,'\n\n')
                        else:
                            _tmp4 = _tmp3[2].rstrip() ##
                            _tmp5 = _tmp4.replace('|','\\\|') + _SCOPE

                            _tmp6 = _tmp5.split(':')
                            if ( len( _tmp5[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexProcess(_process_CUSTOM,_tmp5,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    elif (len(tmp5) >= 2):
                        if (_DEBUG2 == True):
                            print ("ORIG:d:",tmp1[0],"::",_index,sep='',file=fileOUT)
                            print ("_lineMETRIC:f3:",_lineMETRIC,"::",_index,sep='',file=fileOUT)
                            print ("_JMS_PRESENT::",_JMS_PRESENT,"::",_index,sep='',file=fileOUT)


                            _tmp6 = _tmp5.split(':')
                            if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp5[1],_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    elif (len(tmp5) >=1):
                        if (_DEBUG2 == True):
                            print ("ORIG:e:",tmp1[0],"::",_index,sep='',file=fileOUT)
                            print ("_lineMETRIC:f3:",_lineMETRIC,"::",_index,sep='',file=fileOUT)
                            print ("_JMS_PRESENT::",_JMS_PRESENT,"::",_index,sep='',file=fileOUT)
                            print ("tmp5::",tmp5,"::",_index,sep='',file=fileOUT)

                            _tmp6 = tmp5.split(':')
                            if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp5[0],_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    else:
                        print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp5:",len(tmp5),":",tmp5,sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp4:",len(tmp4),":",tmp4,sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp3:",len(tmp3),":",tmp3,sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp1:",len(tmp1),":",tmp1,sep='',file=fileEXCEPTIONS)
                        print("dealing with whitespace, Custom metrics :z:",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)


                    
                else:                           # Regular metrics...
                    if (_DEBUG == True): print("... whitespace, regular", "Got ",len (tmp1)," == ",tmp1," INDEX::",_index,sep='',file=fileOUT)
                        
                    #
                    # assemble the output
                    #
                    tmp3 = tmp1[0].split(_2SPACE)
                    tmp30 = tmp1[0].split(_3SPACE)

                    if (_JMS_PRESENT == True):
                        # get the appropriate units string...
                        for metric in MetricUnit:
                            if ( metric['present'] == True):_tmp1 = _JMS_temp[1].split(metric['metricString'])

                        _tmp2 = _JMS + _tmp1[0]
                        _tmp3 = _tmp2.rstrip()
                        _tmp4 = _tmp3.replace('|','\\\|')
                        
                        if (_DEBUG2 == True):
                            print ("ORIG:f3:",_JMS_temp[1],"::",_index,sep='',file=fileOUT)
                            print ("_lineMETRIC:f3:",_lineMETRIC,"::",_index,sep='',file=fileOUT)
                            print ("_tmp1:f3:",_tmp1,"::",_index,sep='',file=fileOUT)
                            print ("_tmp2:f3:",_tmp2,"::",_index,sep='',file=fileOUT)
                            print ("_tmp3:f3:",_tmp3,"::",_index,sep='',file=fileOUT)
                            print ("_tmp4:f3:",_tmp4,"::",_index,sep='',file=fileOUT)
                            
                        _tmp6 = _tmp4.split(':')
                        if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp4,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    elif ((len(tmp3) >= 6) and (len(tmp3[5]) > 5) ):
                        if (_DEBUG2 == True):print ("ORIG:f0:",tmp1[0],"::",_index,sep='',file=fileOUT)

                        _tmp4 = tmp3[5].replace('|','\\\|')  # alternate delimiter... see above...
                        
                        _tmp6 = _tmp4.split(':')
                        if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp4,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    elif (len(tmp30) >= 3):

                        _tmp4 = tmp30[1].replace('|','\\\|')  # alternate delimiter... see above...
                        _tmp5 = tmp30[2].replace('|','\\\|')  # alternate delimiter... see above...

                        if ( len(tmp30[2]) <5 ):
                            if (_DEBUG2 == True):print ("ORIG:f1:",tmp1[0],"::",_index,sep='',file=fileOUT)

                            _tmp6 = _tmp4.split(':')
                            if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp4,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                        else:
                            if (_DEBUG2 == True):print ("ORIG:f1:",tmp1[0],"::",_index,sep='',file=fileOUT)
                            
                            _tmp6 = _tmp5.split(':')
                            if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                                _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp5,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    elif (len(tmp30) >= 2):
                        _tmp4 = tmp30[1].replace('|','\\\|')  # alternate delimiter... see above...

                        if (_DEBUG2 == True):print ("ORIG:f2:",tmp1[0],"::",_index,sep='',file=fileOUT)

                        _tmp6 = _tmp4.split(':')
                        if ( len( _tmp6[0] ) >5 ):    # a reasonable left hand term...
                            _index, _linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT = writeRegexMetric(_tmp4,_SCOPE,_index,_linesWritten,_metricCOUNT,_validCOUNT,_errorCOUNT)
                    else:
                        print("=====EXCEPTION - parsing failed to yield correct # of tokens ==========",sep='',file=fileEXCEPTIONS)
                        print("TOKENS-tmp3:",len(tmp3),":",tmp3,sep='',file=fileEXCEPTIONS)
                        print("Override with ::",tmp30,sep='',file=fileEXCEPTIONS)
                        print("dealing with regular metrics :zz:",_linesWritten,"\n",sep='',file=fileEXCEPTIONS)
                        
            # reset the metric unit detectors/flags...
            for metric in MetricUnit:
                    metric['present'] = False

            # old hack that needs to be eliminated...
            _MAFE_PRESENT = False
            _MAVG_PRESENT = False
            
            _START_WITH_CUSTOM = False
            _JMS_PRESENT = False
   
    _SYSVIEW_METRIC = False  # reset for next pass
        
print( '\n' )
print( '>>> Evaluated ', _metricCOUNT, ' expressions and got ', _validCOUNT, ' valid with ', _errorCOUNT, ' invalid.' )
print( '>>> Next index is ', _index, ' and ', _linesWritten, ' total lines were written to output file. ','\n\n' )
        
fileOUT.close

